open Absyn
open Format 
open Env 

exception TODO
exception RuntimeError of string

type value = 
    Num of int 
  | ProcVal of env * var list * exp 
  | BoolVal of bool 
  | ConsVal of value * value 
  | NullVal
  | RefVal  of reference 
  | PairVal of value * value
and env = value Env.env
and reference = Reference of int * value array
         
(* --------- values ----------------------------*)
  
let rec print_value (v:value) : unit = 
  match v with 
      Num i -> print_int i;
    | BoolVal b -> 
        if b then print_string "true" else 
          print_string "false"
    | ProcVal (env,v,str) -> print_string "<closure>"
    | RefVal r -> print_string "<reference>"
	 | ConsVal (v1, v2) -> 
		  print_string "cons(";
		  print_value v1; 
		  print_string ","; 
		  print_value v2; 
		  print_string ")"
	 | NullVal -> print_string "[]"
	 | PairVal (v1, v2) -> 
		  print_string "pair(";
		  print_value v1; 
		  print_string ","; 
		  print_value v2; 
		  print_string ")"

let closure (ids:var list) (body:exp) (env:env) : value = 
  ProcVal (env, ids, body)
                
  (* ------------ References and environments ------------- *)

  let deref (Reference (i,vec)) : value = vec.(i) 
  let setref (Reference (i,vec)) (v:value) : unit =  vec.(i) <- v

  (* Unlike the old version of apply_env, when the location of the
	* value is found, that is what is returned as a reference.  *)
  let rec apply_env_ref (env:env) (sym:var) : reference = 
	 match env with 
        Empty -> raise (RuntimeError ("Unbound variable " ^ sym))
		| Extend (vars, vals, old_env) -> 
			 (match list_find_position sym vars with 
					Some pos -> Reference(pos,vals)
				 | None -> apply_env_ref old_env sym)

                                                       
  let extend_env_rec (proc_names:var list) (idss:var list list) (bodies:exp list) (old_env:env) : env = 
    let num_procs = List.length proc_names             in 
    let array     = Array.make num_procs (Num 0)       in
    let new_env   = Extend (proc_names, array,old_env) in 
    let rec loop idss bodies n : unit = 
      match idss, bodies with 
          (ids::is), (body::bs)  -> 
            (array.(n) <- closure ids body new_env;
             loop is bs (n+1))
        | [], [] -> () 
        | _,_ -> failwith "BUG: Cannot happen" in
      begin
        loop idss bodies 0;
        new_env
      end 
                

  (* --------- Booleans --------------- *)
                                                                                                
  let is_true (v:value) :bool = 
    match v with 
        BoolVal b -> b
      | _ -> raise (RuntimeError "Invalid value used as bool")
                         
  let true_value :value  = BoolVal true
  let false_value :value = BoolVal false
 
(* ----------- Primitives ----------- *)

  let rec apply_prim (p:prim) (args:value list) = 
    match p,args with 
        Plus,  [Num v1;Num v2] -> Num (v1 + v2)
      | Times, [Num v1;Num v2] -> Num (v1 * v2)
      | Minus, [Num v1;Num v2] -> Num (v1 - v2)
      | And,   [BoolVal b1; BoolVal b2] -> BoolVal ( b1 && b2 )
      | Or,    [BoolVal b1; BoolVal b2] -> BoolVal (b1 || b2)
      | Not,   [BoolVal b1] -> BoolVal (not b1)
      | Add1, [Num v1] -> Num (v1 + 1)
      | Sub1, [Num v1] -> Num (v1 - 1)
      | Zero, [Num v1] -> if v1 = 0 then true_value else false_value
      | LT, [Num v1; Num v2] -> if v1 < v2 then true_value else false_value
      | GT, [Num v1; Num v2] -> if v1 > v2 then true_value else false_value
      | Equal, [Num v1;Num v2] -> 
          if (v1 = v2) then true_value  else false_value

		| Hd, [ConsVal(v1,v2)] -> v1
		| Tl, [ConsVal(v1,v2)] -> v2
		| Cons, [v1; v2] -> ConsVal (v1,v2)
		| IsNull, [v] -> 
			 (match v with 
					ConsVal (_,_) -> BoolVal true
				 | NullVal -> BoolVal false
				 | _ -> raise (RuntimeError "Incorrect argument for null?"))
			 
		| Pair, [v1; v2] -> PairVal (v1,v2)
		| Fst, [PairVal (v1,v2)] -> v1
		| Snd, [PairVal (v1,v2)] -> v2

      | _, _ -> raise (RuntimeError "Unimplemented  primitive or invalid arguments")


(* ------------ Evaluator ---------------------- *)
                  
  let rec apply_procval (v:value) (args:value list)  : value = 
    match v with 
        ProcVal (env, ids, body) -> 
          if (List.length ids = List.length args) then 
            eval body (extend_env ids args env)
          else 
            raise (RuntimeError "Wrong number of arguments to procedure")
      | _ -> raise (RuntimeError "Attempt to apply non-procedure")

  and  eval (exp:exp) (env:env) :value = 
    match exp with 
        LitInt i -> Num i
      | LitBool b -> 
          if b then true_value 
          else false_value
		| LitNull _ -> NullVal
      | Var v -> apply_env env v
      | Prim (p, rands) -> 
          let args = eval_rands rands env in
            apply_prim p args
      | If (test_exp, true_exp, false_exp) -> 
          if (is_true (eval test_exp env))
          then (eval true_exp env)
          else (eval false_exp env)
      | Let (vars,rands,body) -> 
          let args = eval_rands rands env in 
            eval body (extend_env vars args env)
      | Lambda (tys,ids,body) ->            
          closure ids body env
      | App (rator,rands) -> 
          let proc = eval rator env in 
          let args = eval_rands rands env in 
            apply_procval proc args         
      | Letrec (tys, proc_names, arg_tys, idss, bodies, letrec_body) -> 
          eval letrec_body (extend_env_rec proc_names idss bodies env)
      | Varassign (id, rhs) -> 
          begin
            setref 
              (apply_env_ref env id)
              (eval rhs env);
            Num 1
          end
      | Begin(exp, exps) -> 
          let v = eval exp env in
          let rec loop l = 
            match l with 
                [] -> v
              | [e] -> eval e env 
              | (hd :: tl) -> 
                  ignore (eval hd env) ;
                  loop tl in 
            loop exps
            
  and eval_rands (rands:exp list) (env:env): value list = 
    List.map (fun x -> eval x env) rands

  let eval_program (exp:exp) :unit= 
         print_value (eval exp Empty)
                
